import math

from networkx.utils import py_random_state

from geco.mips.knapsack.generic import knapsack


def _correlated_knapsack_template(
    number_of_items, capacity, profit_generator, weight_generator, profit_first
):
    """A template for generating knapsack instances given a relation between the weights and the profits

    Parameters:
    -----------
    number_of_items: int
        Number of items in knapsack
    capacity: float
        Capacity of knapsack
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance
    """
    profits, weights = generate_from_distribution(
        number_of_items, profit_generator, weight_generator, profit_first
    )
    return knapsack(profits, weights, capacity)


def generate_from_distribution(
    number_of_items, profit_generator, weight_generator, profit_first
):
    if profit_first:
        profits = [profit_generator(None) for _ in range(number_of_items)]
        weights = [weight_generator(profit) for profit in profits]
    else:
        weights = [weight_generator(None) for _ in range(number_of_items)]
        profits = [profit_generator(weight) for weight in weights]

    return profits, weights


@py_random_state(-1)
def uncorrelated_distribution(R, seed=0):
    """
    Generates an uncorrelated distribution as described in section 3 of [1].

    Parameters
    ----------
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return {
        "weight_generator": lambda p: seed.uniform(1, R),
        "profit_generator": lambda w: seed.uniform(1, R),
        "profit_first": True,
    }


@py_random_state(-1)
def uncorrelated(n, c, R=1000, seed=0):
    """
    Generates an uncorrelated distribution knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n, capacity=c, **uncorrelated_distribution(R, seed)
    )


@py_random_state(-1)
def weakly_correlated_distribution(R, seed=0):
    """
    Generates an weakly correlated distribution as described in section 3 of [1].

    Parameters
    ----------
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return {
        "weight_generator": lambda p: seed.uniform(1, R),
        "profit_generator": lambda w: max(1, seed.uniform(w - R / 10, w + R / 10)),
        "profit_first": False,
    }


@py_random_state(-1)
def weakly_correlated(n, c, R=1000, seed=0):
    """
    Generates an weakly correlated distribution knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        **weakly_correlated_distribution(R, seed),
    )


@py_random_state(-1)
def strongly_correlated_distribution(R, seed=0):
    """
    Generates a strongly correlated distribution as described in section 3 of [1].

    Parameters
    ----------
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return {
        "weight_generator": lambda p: seed.uniform(1, R),
        "profit_generator": lambda w: w + R / 10,
        "profit_first": False,
    }


@py_random_state(-1)
def strongly_correlated(n, c, R=1000, seed=0):
    """
    Generates a strongly correlated distribution knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        **strongly_correlated_distribution(R, seed),
    )


@py_random_state(-1)
def inverse_strongly_correlated_distribution(R, seed=0):
    """
    Generates an inverse strongly correlated distribution as described in section 3 of [1].

    Parameters
    ----------
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return {
        "weight_generator": lambda p: p + R / 10,
        "profit_generator": lambda w: seed.uniform(1, R),
        "profit_first": True,
    }


@py_random_state(-1)
def inverse_strongly_correlated(n, c, R=1000, seed=0):
    """
    Generates an inverse strongly correlated distribution knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        **inverse_strongly_correlated_distribution(R, seed),
    )


@py_random_state(-1)
def almost_strongly_correlated_distribution(R, seed=0):
    """
    Generates an almost strongly correlated distribution as described in section 3 of [1].

    Parameters
    ----------
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return {
        "weight_generator": lambda p: seed.uniform(1, R),
        "profit_generator": lambda w: seed.uniform(
            w + R / 10 - R / 500, w + R / 10 - R / 500
        ),
        "profit_first": False,
    }


@py_random_state(-1)
def almost_strongly_correlated(n, c, R=1000, seed=0):
    """
    Generates an almost strongly correlated distribution knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        **almost_strongly_correlated_distribution(R, seed),
    )


@py_random_state(-1)
def subset_sum_distribution(R, seed=0):
    """
    Generates a subset sum distribution as described in section 3 of [1].

    Parameters
    ----------
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return {
        "weight_generator": lambda p: seed.uniform(1, R),
        "profit_generator": lambda w: w,
        "profit_first": False,
    }


@py_random_state(-1)
def subset_sum(n, c, R=1000, seed=0):
    """
    Generates an subset sum distribution knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n, capacity=c, **subset_sum_distribution(R, seed)
    )


@py_random_state(-1)
def uncorrelated_with_similar_weights_distribution(R=None, seed=0):
    """
    Generates an uncorrelated with similar weights distribution as described in section 3 of [1].

    Parameters
    ----------
    R: int
       Unused, kept here to fit the general distribution function prototype
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profit_generator: function
        Takes a weight and generates a profit
    weight_generator: function
        Takes a profit and generates a weight
    profit_first: bool
        Defines if the profits are to be generated first

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return {
        "weight_generator": lambda p: seed.uniform(100_000, 100_100),
        "profit_generator": lambda w: seed.uniform(1, 1000),
        "profit_first": False,
    }


@py_random_state(-1)
def uncorrelated_with_similar_weights(n, c, R=1000, seed=0):
    """
    Generates an uncorrelated with similar weights distribution knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        **uncorrelated_with_similar_weights_distribution(R, seed),
    )


@py_random_state(-1)
def spanner(v, m, n, distribution, capacity, R=1000, seed=0):
    """
    Generates a spanner knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    v: int
        Size of the spanner set
    m: float
        The multiplier limit
    n: int
        Number of items in knapsack
    distribution: dict
        As returned by one of the distribution functions defined in this module
    capacity: float
        Capacity of knapsack
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    # generate the spanner set of v items
    profits, weights = generate_from_distribution(v, **distribution(R, seed))

    # normalize the spanner set
    spanner_profits = [p / m + 1 for p in profits]
    spanner_weights = [w / m + 1 for w in weights]

    # generate n items from spanner set
    profits = []
    weights = []
    spanner_set_indices = list(range(v))
    for _ in range(n):
        idx = seed.choice(spanner_set_indices)
        multiplier = seed.uniform(1, m)
        profit, weight = (
            multiplier * spanner_profits[idx],
            multiplier * spanner_weights[idx],
        )
        profits.append(profit)
        weights.append(weight)

    return knapsack(profits, weights, capacity)


@py_random_state(-1)
def profit_ceiling(n, c, d=3, R=1000, seed=0):
    """
    Generates a profit ceiling knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    d: float
        All profits are multiples of this number
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        profit_generator=lambda w: d * math.ceil(w / d),
        weight_generator=lambda p: seed.uniform(1, R),
        profit_first=False,
    )


@py_random_state(-1)
def circle(n, c, d=2 / 3, R=1000, seed=0):
    """
    Generates a circle knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    d: float
       profit factor
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        profit_generator=lambda w: d * math.sqrt(4 * (R ** 2) - (w - 2 * R) ** 2),
        weight_generator=lambda p: seed.uniform(1, R),
        profit_first=False,
    )


@py_random_state(-1)
def multiple_strongly_correlated(n, c, k1, k2, d, R=1000, seed=0):
    """
    Generates a multiple strongly correlated knapsack instance as described in section 3 of [1].

    Parameters
    ----------
    n: int
        Number of items in knaspack
    c: float
        Capacity of knapsack
    k1: float
        First set profit offset
    k2: float
        Second set profit offset
    d: int
        Weight divisor
    R: int
        Bound for randomization range
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Pisinger. 2005. Where are the hard knapsack problems?
    Comput. Oper. Res. 32, 9 (September 2005), 2271–2284.
    DOI:https://doi.org/10.1016/j.cor.2004.03.002
    """
    return _correlated_knapsack_template(
        number_of_items=n,
        capacity=c,
        profit_generator=lambda w: w + k1 if w % d == 0 else w + k2,
        weight_generator=lambda p: seed.uniform(1, R),
        profit_first=False,
    )



import numpy as np
import scipy
from networkx.utils import np_random_state
import pyscipopt as scip

from geco.mips.set_cover.generic import set_cover


@np_random_state(-1)
def gasse_instance(nrows, ncols, density, max_coef=100, seed=0):
    """
    Generates instance for set cover generation as described in [1].

    Parameters
    ----------
    nrows : int
        Desired number of rows
    ncols : int
        Desired number of columns
    density: float between 0 (excluded) and 1 (included)
        Desired density of the constraint matrix
    max_coef: int
        Maximum objective coefficient (>=1)
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] E.Balas and A.Ho, Set covering algorithms using cutting planes, heuristics,
    and subgradient optimization: A computational study, Mathematical
    Programming, 12 (1980), 37-60.
    """
    return set_cover(
        *gasse_params(nrows, ncols, density, max_coef, seed), name="Gasse Set Cover"
    )


@np_random_state(-1)
def gasse_params(nrows, ncols, density, max_coef=100, seed=0):
    """
    Generates instance params for set cover generation as described in [1],
    based on the code from [2].

    Parameters
    ----------
    nrows : int
        Desired number of rows
    ncols : int
        Desired number of columns
    density: float between 0 (excluded) and 1 (included)
        Desired density of the constraint matrix
    max_coef: int
        Maximum objective coefficient (>=1)
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    costs: list[int]
        Element costs in objective function
    sets: list[set]
        Definition of element requirement for each set

    References
    ----------
    .. [1] E.Balas and A.Ho, Set covering algorithms using cutting planes, heuristics,
    and subgradient optimization: A computational study, Mathematical
    Programming, 12 (1980), 37-60.
    .. [2] https://github.com/ds4dm/learn2branch/blob/master/01_generate_instances.py
    """
    nnzrs = int(nrows * ncols * density)

    assert nnzrs >= nrows  # at least 1 col per row
    assert nnzrs >= 2 * ncols  # at least 2 rows per col

    # compute number of rows per column
    indices = seed.choice(ncols, size=nnzrs)  # random column indexes
    indices[: 2 * ncols] = np.repeat(
        np.arange(ncols), 2
    )  # force at leats 2 rows per col
    _, col_nrows = np.unique(indices, return_counts=True)

    # for each column, sample random rows
    indices[:nrows] = seed.permutation(nrows)  # force at least 1 column per row
    i = 0
    indptr = [0]
    for n in col_nrows:

        # empty column, fill with random rows
        if i >= nrows:
            indices[i : i + n] = seed.choice(nrows, size=n, replace=False)

        # partially filled column, complete with random rows among remaining ones
        elif i + n > nrows:
            remaining_rows = np.setdiff1d(
                np.arange(nrows), indices[i:nrows], assume_unique=True
            )
            indices[nrows : i + n] = seed.choice(
                remaining_rows, size=i + n - nrows, replace=False
            )

        i += n
        indptr.append(i)

    # objective coefficients
    c = seed.randint(max_coef, size=ncols) + 1

    # sparse CSC to sparse CSR matrix
    A = scipy.sparse.csc_matrix(
        (np.ones(len(indices), dtype=int), indices, indptr), shape=(nrows, ncols)
    ).tocsr()
    indices = A.indices
    indptr = A.indptr

    costs = list(c)
    sets = [list(indices[indptr[i] : indptr[i + 1]]) for i in range(nrows)]
    return costs, sets



import pyscipopt as scip


def combinatorial_auction(bids, n_dummy_items, n_items, name="Combinatorial Auction"):
    model = scip.Model(name)

    # add vars
    bids_per_item = [[] for item in range(n_items + n_dummy_items)]
    x = []
    for i, bid in enumerate(bids):
        bundle, price = bid
        var = model.addVar(lb=0, ub=1, obj=price, name=f"x_{i + 1}", vtype="B")
        x.append(var)
        for item in bundle:
            bids_per_item[item].append(i)

    # add constraints
    for item_bids in bids_per_item:
        if item_bids:
            vars = (x[i] for i in item_bids)
            model.addCons(scip.quicksum(vars) <= 1)
    model.setMaximize()
    return model



import networkx as nx
from networkx.utils import py_random_state

from geco.mips.max_cut.generic import naive


@py_random_state(-1)
def tang_instance(n, m, seed=0):
    """
    Generates a max-cut instance as described in A.2 in [1].

    Parameters
    ----------
    n: int
        Number of nodes
    m: int
        Number of edges
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Tang, Y., Agrawal, S., & Faenza, Y. (2019). Reinforcement learning for integer
    programming: Learning to cut. arXiv preprint arXiv:1906.04859.
    """
    graph = nx.generators.gnm_random_graph(n, m, seed=seed)
    weights = tang_params(graph, seed=0)
    for (_, _, data), weight in zip(graph.edges(data=True), weights):
        data["weight"] = weight
    _, model = naive(graph)
    return model


@py_random_state(-1)
def tang_params(graph, seed=0):
    """
    Generates max-cut instance params as described in A.2 in [1].

    Parameters
    ----------
    graph: nx.Graph
        Networkx graph
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    weights: list[int]
        Weight for each edge

    References
    ----------
    .. [1] Tang, Y., Agrawal, S., & Faenza, Y. (2019). Reinforcement learning for integer
    programming: Learning to cut. arXiv preprint arXiv:1906.04859.
    """
    weights = []
    for _ in graph.edges:
        weights.append(seed.randint(0, 10))
    return weights



import itertools

import pytest

from geco.mips.combinatorial_auction.gasse import gasse_params, gasse_instance
from geco.mips.combinatorial_auction.generic import combinatorial_auction


def test_simple_instance():
    bids = [([0], 2)]
    n_dummy_items = 0
    n_items = 1
    instance = combinatorial_auction(
        bids=bids, n_dummy_items=n_dummy_items, n_items=n_items
    )
    instance.hideOutput()
    instance.optimize()
    assert instance.getObjectiveSense() == "maximize"
    assert instance.getNVars() == len(bids)
    assert instance.getNConss() == n_items
    assert instance.getStatus() == "optimal"
    assert instance.getObjVal() == 2  # optimal value is the price of the only bid


def test_gasse_instance_creation():
    n_bids = 1
    n_items = 1
    model = gasse_instance(
        n_items=n_items,
        n_bids=n_bids,
        min_value=1,
        max_value=2,
        value_deviation=0,
        add_item_prob=1,
        max_n_sub_bids=1,
        seed=0,
    )
    assert model.getObjectiveSense() == "maximize"
    assert model.getNVars() == n_bids
    assert model.getNConss() == n_items


@pytest.mark.parametrize(
    "n_items, n_bids, min_value, max_value, value_deviation, add_item_prob, max_n_sub_bids, additivity, "
    "budget_factor, resale_factor, integers, warnings, seed1, seed2",
    itertools.product(
        [5, 10, 100],
        [500],
        [1],
        [100],
        [-0.5, 0.5],
        [0.9],
        [5],
        [0.2],
        [1.5],
        [0.5, 1],
        [True, False],
        [True, False],
        [1, 1337],
        [1, 1337],
    ),
)
def test_gasse_seeding(
    n_items,
    n_bids,
    min_value,
    max_value,
    value_deviation,
    add_item_prob,
    max_n_sub_bids,
    additivity,
    budget_factor,
    resale_factor,
    integers,
    warnings,
    seed1,
    seed2,
):
    params1 = gasse_params(
        n_items,
        n_bids,
        min_value,
        max_value,
        value_deviation,
        add_item_prob,
        max_n_sub_bids,
        additivity,
        budget_factor,
        resale_factor,
        integers,
        warnings,
        seed=seed1,
    )
    params2 = gasse_params(
        n_items,
        n_bids,
        min_value,
        max_value,
        value_deviation,
        add_item_prob,
        max_n_sub_bids,
        additivity,
        budget_factor,
        resale_factor,
        integers,
        warnings,
        seed=seed2,
    )
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params



import itertools

from networkx.utils import py_random_state
from pyscipopt import scip


def late_tasks_formulation(
    number_of_facilities,
    number_of_tasks,
    time_steps,
    processing_times,
    capacities,
    assignment_costs,
    release_dates,
    deadlines,
    name="Hooker Scheduling Late Tasks Formulation",
):
    """Generates late tasks mip formulation described in section 4 in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    time_steps:
        the number of time steps starting from 0 (corresponds to "N" in the paper)
    processing_times: dict[int,int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[int,int]
        cost of assigning a task to a facility
    release_dates: list[int]
        time step at which a job is released
    deadlines: dict[int, float]
        deadline (time step) to finish a job
    name: str
        assigned name to generated instance

    Returns
    -------
        model: SCIP model of the late tasks instance

    References
    ----------
    .. [1] Hooker, John. (2005). Planning and Scheduling to Minimize
     Tardiness. 314-327. 10.1007/11564751_25.
    """
    model = scip.Model(name)

    start_time = min(release_dates)
    time_steps = range(start_time, start_time + time_steps)

    # add variables and their cost
    L = []
    for i in range(number_of_tasks):
        var = model.addVar(lb=0, ub=1, obj=1, name=f"L_{i}", vtype="B")
        L.append(var)

    # assignment vars
    x = {}
    for j, i, t in itertools.product(
        range(number_of_tasks), range(number_of_facilities), time_steps
    ):
        var = model.addVar(lb=0, ub=1, obj=0, name=f"x_{j}_{i}_{t}", vtype="B")
        x[j, i, t] = var

    # add constraints
    # constraint (a)
    for j, t in itertools.product(range(number_of_tasks), time_steps):
        model.addCons(
            len(time_steps) * L[j]
            >= scip.quicksum(
                (
                    (t + processing_times[j, i]) * x[j, i, t] - deadlines[j]
                    for i in range(number_of_facilities)
                )
            )
        )

    # constraint (b)
    for j in range(number_of_tasks):
        vars = (
            x[j, i, t]
            for i, t in itertools.product(range(number_of_facilities), time_steps)
        )
        model.addCons(scip.quicksum(vars) == 1)

    # constraint (c)
    for i, t in itertools.product(range(number_of_facilities), time_steps):
        vars = []
        for j in range(number_of_tasks):
            vars += [
                assignment_costs[j, i] * x[j, i, t_prime]
                for t_prime in range(t - processing_times[j, i] + 1, t + 1)
                if (j, i, t_prime) in x
            ]
        model.addCons(scip.quicksum(vars) <= capacities[i])

    # constraint (d)
    for i, j, t in itertools.product(
        range(number_of_facilities), range(number_of_tasks), time_steps
    ):
        if t < release_dates[j] or t > len(time_steps) - processing_times[j, i]:
            model.addCons(x[j, i, t] == 0)

    model.setMinimize()

    return model


def heinz_formulation(
    number_of_facilities,
    number_of_tasks,
    processing_times,
    capacities,
    assignment_costs,
    release_dates,
    deadlines,
    resource_requirements,
    name="Heinz Scheduling Formulation",
):
    """Generates scheduling MIP formulation according to Model 4 in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    processing_times: dict[(int,int),int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[(int,int),int]
        cost of assigning a task to a facility
    release_dates: list[int]
        time step at which a job is released
    deadlines: dict[int, float]
        deadline (time step) to finish a job
    resource_requirements: dict[(int,int),int]
        resources required for each task assigned to a facility
    name: str
        assigned name to generated instance

    Returns
    -------
        model: SCIP model of generated instance

    References
    ----------
    .. [1] Heinz, J. (2013). Recent Improvements Using Constraint Integer Programming for Resource Allocation and Scheduling.
    In Integration of AI and OR Techniques in Constraint Programming for Combinatorial Optimization Problems
    (pp. 12–27). Springer Berlin Heidelberg.
    """
    model = scip.Model(name)

    time_steps = range(min(release_dates), int(max(deadlines)))

    # objective function
    x = {}
    for j, k in itertools.product(range(number_of_tasks), range(number_of_facilities)):
        var = model.addVar(
            lb=0, ub=1, obj=assignment_costs[j, k], name=f"x_{j}_{k}", vtype="B"
        )
        x[j, k] = var

    # y vars
    y = {}
    for j, k, t in itertools.product(
        range(number_of_tasks), range(number_of_facilities), time_steps
    ):
        if release_dates[j] <= t <= deadlines[j] - processing_times[j, k]:
            var = model.addVar(lb=0, ub=1, obj=0, name=f"y_{j}_{k}_{t}", vtype="B")
            y[j, k, t] = var

    # add constraints
    # constraint (12)
    for j in range(number_of_tasks):
        model.addCons(scip.quicksum(x[j, k] for k in range(number_of_facilities)) == 1)

    # constraint (13)
    for j, k in itertools.product(range(number_of_tasks), range(number_of_facilities)):
        model.addCons(
            scip.quicksum(
                y[j, k, t]
                for t in range(
                    release_dates[j], int(deadlines[j]) - processing_times[j, k]
                )
                if t < len(time_steps)
            )
            == x[j, k]
        )

    # constraint (14)
    for k, t in itertools.product(range(number_of_facilities), time_steps):
        model.addCons(
            scip.quicksum(
                resource_requirements[j, k] * y[j, k, t_prime]
                for j in range(number_of_tasks)
                for t_prime in range(t - processing_times[j, k], t + 1)
                if (j, k, t_prime) in y
            )
            <= capacities[k]
        )

    # constraint (15)
    epsilon = filter(
        lambda ts: ts[0] < ts[1], itertools.product(release_dates, deadlines)
    )
    for k, (t1, t2) in itertools.product(range(number_of_facilities), epsilon):
        model.addCons(
            scip.quicksum(
                processing_times[j, k] * resource_requirements[j, k] * x[j, k]
                for j in range(number_of_tasks)
                if t1 <= release_dates[j] and t2 >= deadlines[j]
            )
            <= capacities[k] * (t2 - t1)
        )

    return model


def hooker_cost_formulation(
    number_of_facilities,
    number_of_tasks,
    processing_times,
    capacities,
    assignment_costs,
    release_dates,
    deadlines,
    resource_requirements,
    name="Hooker Cost Scheduling Formulation",
):
    """Generates scheduling MIP formulation according to [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    processing_times: dict[(int,int),int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[(int,int),int]
        cost of assigning a task to a facility
    release_dates: list[int]
        time step at which a job is released
    deadlines: dict[int, float]
        deadline (time step) to finish a job
    resource_requirements: dict[(int,int),int]
        resources required for each task assigned to a facility
    name: str
        assigned name to generated instance

    Returns
    -------
        model: SCIP model of generated instance

    References
    ----------
    .. [1] J. N. Hooker, A hybrid method for planning and scheduling, CP 2004.
    """
    model = scip.Model(name)

    time_steps = range(min(release_dates), int(max(deadlines)))

    # objective function
    x = {}
    for j, i, t in itertools.product(
        range(number_of_tasks), range(number_of_facilities), time_steps
    ):
        var = model.addVar(
            lb=0, ub=1, obj=assignment_costs[j, i], name=f"x_{j}_{i}_{t}", vtype="B"
        )
        x[j, i, t] = var

    # add constraints
    # constraints (a)
    for j in range(number_of_tasks):
        model.addCons(
            scip.quicksum(
                x[j, i, t]
                for i, t in itertools.product(range(number_of_facilities), time_steps)
            )
            == 1
        )

    # constraints (b)
    for i, t in itertools.product(range(number_of_facilities), time_steps):
        model.addCons(
            scip.quicksum(
                resource_requirements[j, i] * x[j, i, t] for j in range(number_of_tasks)
            )
            <= capacities[i]
        )

    # constraints (c)
    for j, i, t in itertools.product(
        range(number_of_tasks), range(number_of_facilities), time_steps
    ):
        if (
            deadlines[j] - processing_times[j, i] < t < release_dates[j]
            or t > number_of_tasks - processing_times[j, i]
        ):
            model.addCons(x[j, i, t] == 0)

    return model


@py_random_state(-1)
def generate_params(number_of_facilities, number_of_tasks, seed=0):
    """
    Generic instance parameter generator for heinz [1] and hooker [2] formulations.
    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization
    Returns
    -------
    processing_times: dict[int,int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[int,int]
        cost of assigning a task to a facility
    release_times: list[int]
        time step at which a job is released
    deadlines: dict[int, float]
        deadline (time step) to finish a job
    resource_requirements: dict[int,int]
        resources required for each task assigned to a facility
    References
    ----------
    .. [1] Heinz, J. (2013). Recent Improvements Using Constraint Integer Programming for Resource Allocation and Scheduling.
    In Integration of AI and OR Techniques in Constraint Programming for Combinatorial Optimization Problems
    (pp. 12–27). Springer Berlin Heidelberg.
    .. [2] Hooker, John. (2005). Planning and Scheduling to Minimize
     Tardiness. 314-327. 10.1007/11564751_25.
    """
    processing_times = {}

    for j, i in itertools.product(range(number_of_tasks), range(number_of_facilities)):
        if number_of_tasks < 22:
            processing_times[j, i] = seed.randint(2, 20 + 5 * i)
        else:
            processing_times[j, i] = seed.randint(5, 20 + 5 * i)

    capacities = [10] * number_of_facilities

    assignment_costs = {}
    for j in range(number_of_tasks):
        value = seed.randint(1, 10)
        for i in range(number_of_facilities):
            assignment_costs[j, i] = value

    release_times = [0] * number_of_tasks

    beta = 20 / 9
    deadlines = [
        seed.uniform(beta * number_of_tasks / 4, beta * number_of_tasks)
        for _ in range(number_of_tasks)
    ]

    resource_requirements = {}
    for j, k in itertools.product(range(number_of_tasks), range(number_of_facilities)):
        resource_requirements[j, k] = seed.randint(1, 9)

    return (
        processing_times,
        capacities,
        assignment_costs,
        release_times,
        deadlines,
        resource_requirements,
    )
